/*
* code by jennkirby
* https://www.instructables.com/Wireless-Gametrak/
* CC BY-NC-SA 4.0
* makes use of examples from adafruit libraries
*/

#include <WiFi.h>
#include <WiFiUdp.h>
#include "arduino_secrets.h" 
#include <Adafruit_TCA8418.h>
#include <Adafruit_LSM6DSOX.h>
#include <Adafruit_LIS3MDL.h>

Adafruit_LSM6DSOX lsm6ds;
Adafruit_LIS3MDL lis3mdl;

const char* ssid = "YOUR ROUTER NAME"; //TODO: update to your router name
const char* password = "YOUR ROUTER PASSWORD"; //TODO: update to your router password

WiFiUDP Udp;
IPAddress staticIP(192, 168, 1, 101); //TODO: update to unique IP on the network (you will need to change this per wireless gametrak)
IPAddress gateway(192, 168, 1, 1);
IPAddress subnet(255, 255, 0, 0);
unsigned int localUdpPort = 4222;  // TODO: local port to listen on (must be unique)
unsigned int outUdpPort = 4444;  // TODO: local port to listen on (must be unique)
char incomingPacket[255];  // buffer for incoming packets
char replyPacket[] = "got message";  // a reply string to send back

int led = LED_BUILTIN;

//use ADC1 when using WIFI. ADC2 not available on ESP32
#define T1XPIN A2
#define T1YPIN A3
#define T1ZPIN A4
#define T2XPIN 33
#define T2YPIN 32
#define T2ZPIN 37

#define VBATPIN A13


int t1X;//green
int t1Y;//yellow
int t1Z;//orange
int t2X;//green
int t2Y;//yellow
int t2Z;//orange
//blue ground
//red 3v

//keypad
Adafruit_TCA8418 keypad;
#define ROWS 4
#define COLS 4

int keymap[COLS][ROWS] = {
    {15, 0, 14, 13},
    {7, 8, 9, 12},
    {4, 5, 6, 11},
    {1, 2, 3, 10},
};

int aKey = -1;

char printStr[255]; //array for printing values

void setup() {
  Serial.begin(115200);

  // set LED to be an output pin and turn on 
  pinMode(led, OUTPUT);
  digitalWrite(led, HIGH);

  keypadSetup();

  imuSetup();

  if (WiFi.config(staticIP, gateway, subnet) == false) {
    Serial.println("Configuration failed.");
  }
  Serial.printf("Connecting to %s ", ssid);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED)
  {
    delay(500);
    Serial.print(".");
  }
  Serial.println(" connected");

  Udp.begin(localUdpPort);
  Serial.printf("Now listening at IP %s, UDP port %d\n", WiFi.localIP().toString().c_str(), localUdpPort);
}

void loop() {
  t1X = analogRead(T1XPIN);
  t1Y = analogRead(T1YPIN);
  t1Z = analogRead(T1ZPIN);
  t2X = analogRead(T2XPIN);
  t2Y = analogRead(T2YPIN);
  t2Z = analogRead(T2ZPIN);

  int measuredvbat = analogReadMilliVolts(VBATPIN);
  measuredvbat *= 2;    // we divided by 2, so multiply back
  measuredvbat /= 100; // convert to volts - should divide by 1000, dividing by 100 to send as int over sprintf, divide by 10 on other side. Should max out around 4.2 (42)
  //Serial.print("VBat: " ); Serial.println(measuredvbat);

  if (keypad.available() > 0) {
    int k = keypad.getEvent();
    bool pressed = k & 0x80;
    k &= 0x7F;
    k--;
    uint8_t row = k / 10;
    uint8_t col = k % 10;
    if (pressed)
      aKey = keymap[col][row];
    else
      aKey = -1;
  }

  //read IMU
  sensors_event_t accel, gyro, mag, temp;
  //  /* Get new normalized sensor events */
  lsm6ds.getEvent(&accel, &gyro, &temp);
  lis3mdl.getEvent(&mag);

  //build string of all data
  sprintf(printStr, "t1X %d t1Y %d t1Z %d t2X %d t2Y %d t2Z %d key %d Accel %.2f %.2f %.2f Mag %.2f %.2f %.2f Gyro %.2f %.2f %.2f bat %d", t1X, t1Y, t1Z, t2X, t2Y, t2Z, aKey,  accel.acceleration.x, accel.acceleration.y, accel.acceleration.z, gyro.gyro.x, gyro.gyro.y, gyro.gyro.z, mag.magnetic.x, mag.magnetic.y, mag.magnetic.z, measuredvbat);
  //Serial.println(printStr);

  int packetSize = Udp.parsePacket();
  if (packetSize)
  {
    // receive incoming UDP packets
    Serial.printf("Received %d bytes from %s, port %d\n", packetSize, Udp.remoteIP().toString().c_str(), Udp.remotePort());
    int len = Udp.read(incomingPacket, 255);
    if (len > 0)
    {
      incomingPacket[len] = 0;
    }
    Serial.printf("UDP packet contents: %s\n", incomingPacket);
  }

  // send back a reply, to the IP address and port we got the packet from
  Udp.beginPacket(Udp.remoteIP(), outUdpPort);

  //int i = 0;
  //while (printStr[i] != 0) 
    //Udp.write((uint8_t)printStr[i++]);
  Udp.print(printStr);
  Udp.endPacket();

  delay(20);  
}
void keypadSetup() {
  if (!keypad.begin(TCA8418_DEFAULT_ADDR, &Wire)) {
    Serial.println("keypad not found, check wiring & pullups!");
  }
  // configure the size of the keypad matrix.
  // all other pins will be inputs
  keypad.matrix(ROWS, COLS);

  // flush the internal buffer
  keypad.flush();
}
void imuSetup() {
  bool lsm6ds_success, lis3mdl_success;
  // hardware I2C mode, can pass in address & alt Wire
  lsm6ds_success = lsm6ds.begin_I2C();
  lis3mdl_success = lis3mdl.begin_I2C();

  if (!lsm6ds_success){
    Serial.println("Failed to find LSM6DS chip");
  }
  if (!lis3mdl_success){
    Serial.println("Failed to find LIS3MDL chip");
  }
  if (lsm6ds_success && lis3mdl_success){
    Serial.println("LSM6DS and LIS3MDL Found!");

    // lsm6ds.setAccelRange(LSM6DS_ACCEL_RANGE_2_G);
    Serial.print("Accelerometer range set to: ");
    switch (lsm6ds.getAccelRange()) {
    case LSM6DS_ACCEL_RANGE_2_G:
      Serial.println("+-2G");
      break;
    case LSM6DS_ACCEL_RANGE_4_G:
      Serial.println("+-4G");
      break;
    case LSM6DS_ACCEL_RANGE_8_G:
      Serial.println("+-8G");
      break;
    case LSM6DS_ACCEL_RANGE_16_G:
      Serial.println("+-16G");
      break;
    }

    // lsm6ds.setAccelDataRate(LSM6DS_RATE_12_5_HZ);
    Serial.print("Accelerometer data rate set to: ");
    switch (lsm6ds.getAccelDataRate()) {
    case LSM6DS_RATE_SHUTDOWN:
      Serial.println("0 Hz");
      break;
    case LSM6DS_RATE_12_5_HZ:
      Serial.println("12.5 Hz");
      break;
    case LSM6DS_RATE_26_HZ:
      Serial.println("26 Hz");
      break;
    case LSM6DS_RATE_52_HZ:
      Serial.println("52 Hz");
      break;
    case LSM6DS_RATE_104_HZ:
      Serial.println("104 Hz");
      break;
    case LSM6DS_RATE_208_HZ:
      Serial.println("208 Hz");
      break;
    case LSM6DS_RATE_416_HZ:
      Serial.println("416 Hz");
      break;
    case LSM6DS_RATE_833_HZ:
      Serial.println("833 Hz");
      break;
    case LSM6DS_RATE_1_66K_HZ:
      Serial.println("1.66 KHz");
      break;
    case LSM6DS_RATE_3_33K_HZ:
      Serial.println("3.33 KHz");
      break;
    case LSM6DS_RATE_6_66K_HZ:
      Serial.println("6.66 KHz");
      break;
    }

    // lsm6ds.setGyroRange(LSM6DS_GYRO_RANGE_250_DPS );
    Serial.print("Gyro range set to: ");
    switch (lsm6ds.getGyroRange()) {
    case LSM6DS_GYRO_RANGE_125_DPS:
      Serial.println("125 degrees/s");
      break;
    case LSM6DS_GYRO_RANGE_250_DPS:
      Serial.println("250 degrees/s");
      break;
    case LSM6DS_GYRO_RANGE_500_DPS:
      Serial.println("500 degrees/s");
      break;
    case LSM6DS_GYRO_RANGE_1000_DPS:
      Serial.println("1000 degrees/s");
      break;
    case LSM6DS_GYRO_RANGE_2000_DPS:
      Serial.println("2000 degrees/s");
      break;
    case ISM330DHCX_GYRO_RANGE_4000_DPS:
      Serial.println("4000 degrees/s");
      break;
    }
    // lsm6ds.setGyroDataRate(LSM6DS_RATE_12_5_HZ);
    Serial.print("Gyro data rate set to: ");
    switch (lsm6ds.getGyroDataRate()) {
    case LSM6DS_RATE_SHUTDOWN:
      Serial.println("0 Hz");
      break;
    case LSM6DS_RATE_12_5_HZ:
      Serial.println("12.5 Hz");
      break;
    case LSM6DS_RATE_26_HZ:
      Serial.println("26 Hz");
      break;
    case LSM6DS_RATE_52_HZ:
      Serial.println("52 Hz");
      break;
    case LSM6DS_RATE_104_HZ:
      Serial.println("104 Hz");
      break;
    case LSM6DS_RATE_208_HZ:
      Serial.println("208 Hz");
      break;
    case LSM6DS_RATE_416_HZ:
      Serial.println("416 Hz");
      break;
    case LSM6DS_RATE_833_HZ:
      Serial.println("833 Hz");
      break;
    case LSM6DS_RATE_1_66K_HZ:
      Serial.println("1.66 KHz");
      break;
    case LSM6DS_RATE_3_33K_HZ:
      Serial.println("3.33 KHz");
      break;
    case LSM6DS_RATE_6_66K_HZ:
      Serial.println("6.66 KHz");
      break;
    }

    lis3mdl.setDataRate(LIS3MDL_DATARATE_155_HZ);
    // You can check the datarate by looking at the frequency of the DRDY pin
    Serial.print("Magnetometer data rate set to: ");
    switch (lis3mdl.getDataRate()) {
      case LIS3MDL_DATARATE_0_625_HZ: Serial.println("0.625 Hz"); break;
      case LIS3MDL_DATARATE_1_25_HZ: Serial.println("1.25 Hz"); break;
      case LIS3MDL_DATARATE_2_5_HZ: Serial.println("2.5 Hz"); break;
      case LIS3MDL_DATARATE_5_HZ: Serial.println("5 Hz"); break;
      case LIS3MDL_DATARATE_10_HZ: Serial.println("10 Hz"); break;
      case LIS3MDL_DATARATE_20_HZ: Serial.println("20 Hz"); break;
      case LIS3MDL_DATARATE_40_HZ: Serial.println("40 Hz"); break;
      case LIS3MDL_DATARATE_80_HZ: Serial.println("80 Hz"); break;
      case LIS3MDL_DATARATE_155_HZ: Serial.println("155 Hz"); break;
      case LIS3MDL_DATARATE_300_HZ: Serial.println("300 Hz"); break;
      case LIS3MDL_DATARATE_560_HZ: Serial.println("560 Hz"); break;
      case LIS3MDL_DATARATE_1000_HZ: Serial.println("1000 Hz"); break;
    }

    lis3mdl.setRange(LIS3MDL_RANGE_4_GAUSS);
    Serial.print("Range set to: ");
    switch (lis3mdl.getRange()) {
      case LIS3MDL_RANGE_4_GAUSS: Serial.println("+-4 gauss"); break;
      case LIS3MDL_RANGE_8_GAUSS: Serial.println("+-8 gauss"); break;
      case LIS3MDL_RANGE_12_GAUSS: Serial.println("+-12 gauss"); break;
      case LIS3MDL_RANGE_16_GAUSS: Serial.println("+-16 gauss"); break;
    }

    lis3mdl.setPerformanceMode(LIS3MDL_MEDIUMMODE);
    Serial.print("Magnetometer performance mode set to: ");
    switch (lis3mdl.getPerformanceMode()) {
      case LIS3MDL_LOWPOWERMODE: Serial.println("Low"); break;
      case LIS3MDL_MEDIUMMODE: Serial.println("Medium"); break;
      case LIS3MDL_HIGHMODE: Serial.println("High"); break;
      case LIS3MDL_ULTRAHIGHMODE: Serial.println("Ultra-High"); break;
    }

    lis3mdl.setOperationMode(LIS3MDL_CONTINUOUSMODE);
    Serial.print("Magnetometer operation mode set to: ");
    // Single shot mode will complete conversion and go into power down
    switch (lis3mdl.getOperationMode()) {
      case LIS3MDL_CONTINUOUSMODE: Serial.println("Continuous"); break;
      case LIS3MDL_SINGLEMODE: Serial.println("Single mode"); break;
      case LIS3MDL_POWERDOWNMODE: Serial.println("Power-down"); break;
    }

    lis3mdl.setIntThreshold(500);
    lis3mdl.configInterrupt(false, false, true, // enable z axis
                            true, // polarity
                            false, // don't latch
                            true); // enabled!
  }
}
